/*
 * cmd_rtk_otp.c - interface to One-Time-Programmable memory
 *
 * Copyright (c) 2008-2015 Analog Devices Inc.
 *
 * Licensed under the GPL-2 or later.
 */
#include <common.h>
#include <command.h>
#include <malloc.h>
#include <net.h>
#include <drivers/otp/rtk_otp_util.h>
#include <drivers/otp/rtk_otp_burn.h>
#include <drivers/otp/rtk_otp_ecc_algorithm.h>

#define DOWNLOAD_ADDR 0x108000

typedef enum {
    OTP_CMD_HELP,
    OTP_CMD_DUMP,
    OTP_CMD_PROGRAM,
    OTP_CMD_BITMAP_PROGRAM,
    OTP_CMD_CRC_AUTO_REGION,
    OTP_CMD_CRC_PARTIAL_REGION,
    OTP_CMD_QUERY_EVER_PROGRAMED,
    OTP_CMD_ECC_CHECK,
} OTP_CMD_ID;

typedef struct {
    OTP_CMD_ID      id;
    unsigned long   flags;

#define VALUE_DETECTED     (1<<1)

    unsigned long   ofst;
    unsigned long   msb;
    unsigned long   lsb;
    unsigned long   val;
    unsigned long start_ofst;
    unsigned long end_ofst;
    unsigned long ecc_ofst;
    unsigned long ecc_unit;
    unsigned long ignore_ecc_not_burn;
    void *data;
    int len;
} OTP_CMD;

static int str_to_long(char *str, unsigned long* p_val, unsigned char base)
{
    char *p;
    long int  val;

    base = (base == 8 || base == 10 || base == 16 || base == 32) ? base : 0;
    val = simple_strtoul(str, &p, base);

    if (*p == '\0') {
        *p_val = val;
        return 0;
    }
    return -1;
}

static int down_bitmap_file(char *fileName)
{
    int size;

    if(!fileName)
        return 0;
    load_addr = DOWNLOAD_ADDR;
    copy_filename(BootFile, fileName, sizeof(BootFile));
    if ((size = NetLoop(TFTPGET)) < 0) {
        return 0;
    }
    /* done if no file was loaded (no errors though) */
    if (size == 0) {
        return 0;
    }
    /* flush cache */
    flush_cache(DOWNLOAD_ADDR, size);
    return size;
}

/*******************************************************************************
 * FUNC : do_otp_dump
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_dump(unsigned long addr, unsigned char msb, unsigned char lsb)
{
    unsigned long val;

    if (otp_read_bits(addr, msb, lsb, &val) < 0)
        printf("dump OTP[0x%lx] bits[%d:%d] failed\n", addr, msb, lsb);
    else {
        if (val == 0xdeaddead)
            printf("OTP[0x%lx] = 0x%lx (this OTP is non readable)\n", addr, val);
        else
            printf("OTP[0x%lx][%d:%d] = 0x%lx\n", addr, msb, lsb, val);
    }
}

/*******************************************************************************
 * FUNC : do_otp_dump
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_read_region(unsigned long addr, unsigned int len)
{
    unsigned int *data = NULL;
    if (len < 0) {
        return;
    }
    data = (unsigned int *)malloc(sizeof(unsigned int) * len);
    if(data == NULL)
        return;
    memset(data, 0, sizeof(unsigned int) * len);
    if (otp_read_region(addr, data, len) != 0) {
        printf("dump OTP region failed\n");
    } else {
        printf("the otp region[0x%lx-0x%lx] is: 0x", addr, addr + sizeof(unsigned int) * (len - 1));
        while(len > 0) {
            printf("%08x", data[len - 1]);
            len --;
        }
        printf("\n");
    }
    free(data);
    data = NULL;
}

/*******************************************************************************
 * FUNC : do_otp_program
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_program(unsigned long addr, unsigned char msb, unsigned char lsb, unsigned long val)
{
    int ret = 0;
    otp_prepare_burn();
    ret = otp_write_bits(addr, msb, lsb, val);
    otp_post_burn();
    if (ret < 0) {
        printf("program OTP[0x%lx] bits[%d:%d] failed\n", addr, msb, lsb);
    } else {
        printf("program OTP[0x%lx][%d:%d] to 0x%lx successed\n", addr, msb, lsb, val);
        do_otp_dump(addr, msb, lsb);
    }

}



/*******************************************************************************
 * FUNC : do_otp_program_region
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static int do_otp_program_region(
    unsigned long addr,
    unsigned int *data,
    int len
)
{
    int ret = 0;
    otp_prepare_burn();
    ret = otp_program_region(addr, data, len);
    otp_post_burn();
    if(ret == 0)
        printf("program OTP sucess, ofst = 0x%08lx, len = %08x\n", addr, len);
    else
        printf("program OTP failed, ofst = 0x%08lx, len = %08x\n", addr, len);
    
    return ret;
}


/*******************************************************************************
 * FUNC : do_otp_program_with_bitmap
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static int do_otp_program_with_bitmap(void)
{
    int ret = 0;
    if(!otp_is_ever_programed())
        ret = otp_program_with_bitmap_with_verify();
    else
        ret = otp_program_with_bitmap();
    return ret;
}


/*******************************************************************************
 * FUNC : do_otp_crc_auto_region
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_crc_auto_region(void)
{
    unsigned long val;
    int ret = -1;
    ret = otp_crc_auto_region(&val);
    if (ret == 0)
        printf("OTP CRC AUTO REGION SUCESS,the  crc auto region value = 0x%08lx\n", val);
    else
        printf("OTP CRC AUTO REGION FAILED\n");
}

/*******************************************************************************
 * FUNC : do_otp_crc_partial_region
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_crc_partial_region(unsigned long start_ofst, unsigned long end_ofst)
{
    unsigned long val;
    int ret = -1;
    ret = otp_crc_partial_region(start_ofst, end_ofst, &val);
    if (ret == 0)
        printf("OTP CRC PARTIAL REGION SUCESS,the  crc partial region value = 0x%08lx\n", val);
    else
        printf("OTP CRC PARTIAL REGION FAILED\n");
}

/*******************************************************************************
 * FUNC : do_otp_crc_partial_region
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static int do_query_otp_ever_programed(void)
{
    int ret = otp_is_ever_programed();
    if(ret == 1)
        printf("OTP has ever been programed\n");
    else
        printf("OTP has not been programed\n");
    return ret;
}

/*******************************************************************************
 * FUNC : do_otp_get_data_with_ecc_check
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_otp_get_data_with_ecc_check(
    unsigned int data_ofst, unsigned int data_len,
    unsigned int ecc_start_bit, unsigned int ecc_unit_bit_num,
    unsigned int ignore_ecc_not_burn)
{
    printf("This IC don't support this function\n");
}


/*******************************************************************************
 * FUNC : do_help
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static void do_help(void)
{
    printf("                                                    \n");
    printf(" @@@ Realtek OTP Tool @@@                           \n");
    printf("                                                    \n");
    printf(" Format1 : otp_util <opcode> ofst=<ofst> bits=<msb>:<lsb> [val=<otp value>] [bitmap_file=<bitmap file>]\n");
    printf("                                                    \n");
    printf(" Format2 : otp_util <opcode> start_ofst=<start_ofst>[end_ofst=<end_ofst>][val=<otp value>][len=<len>]\n");
    printf("                                                    \n");
    printf("   <opcode> : operations                            \n");
    printf("      -d  : dump otp                                \n");
    printf("      -p  : program otp                             \n");
    printf("      -b  : program otp with otp_map.bin          \n");
    printf("      -c  : crc auto region                             \n");
    printf("      -e  : crc partial region                             \n");
    printf("      -h  : print help                              \n");
    printf("                                                    \n");
    printf("   <ofst> : otp offset in byte, should be 4 byte align)\n");
    printf("            presented in hex format, and should be multiple of 4\n");
    printf("                                                    \n");
    printf("   <msb>  : most significant bits (0~31,default: 31)\n");
    printf("                                                    \n");
    printf("   <lsb>  : last significant bits (0~31,default: 0) \n");
    printf("                                                    \n");
    printf("   <bitmap file>  : file that contrain otp bits     \n");
    printf("                                                    \n");
    printf("   <start_ofst>  : the offset of the start reg of the reg region you want to do crc verify     \n");
    printf("                                                    \n");
    printf("   <end_ofst>  : the offset of the last reg of the reg region you want to do crc verify     \n");
    printf("                                                    \n");
    printf("   <len>  : the number of regs of the reg region you want to do crc verify     \n");
    printf("                                                    \n");
    printf(" Example: Dump OTP bits                             \n");
    printf("      otp_util -d ofst=<ofst> bits=<msb>:<lsb>    \n");
    printf("                                                    \n");
    printf(" Example: Dump OTP region                             \n");
    printf("      otp_util -d ofst=<ofst> len=<len>    \n");
    printf("                                                    \n");
    printf(" Example: Program OTP bits (direct assign value)    \n");
    printf("      otp_util -p ofst=<ofst> bits=<msb>:<lsb> val=<val (MSB first)> \n");
    printf("                                                    \n");
    printf(" Example: Program OTP bits (the data should be contained in DDR)      \n");
    printf("      otp_util -p ofst=<ofst> data=<data addr> len=<len>     \n");
    printf("                                                    \n");
    printf(" Example: Program OTP bits with default otp_map.bin      \n");
    printf("      otp_util -b     \n");
    printf("                                                    \n");
    printf(" Example: Program OTP bits use bitmap file downloaded from tftp server.      \n");
    printf("               Note:before run this command,plese set tftp parameters.     \n");
    printf("                                                    \n");
    printf("      otp_util -b [ofst=<ofst>] bitmap_file=<bitmap_file>       \n");
    printf("                                                    \n");
    printf(" Example: crc auto region      \n");
    printf("      otp_util -c       \n");
    printf("                                                    \n");
    printf(" Example: crc partial region      \n");
    printf("      otp_util -e start_ofst=<ofst> end_ofst=<ofst>      \n");
    printf("                                                    \n");
    printf(" Example: crc partial region      \n");
    printf("      otp_util -e start_ofst=<ofst> len=<len>      \n");
    printf("                                                    \n");
    printf(" Example: query whether OTP has ever been programed      \n");
    printf("      otp_util -q      \n");
    printf("                                                    \n");
    printf(" Example: ecc check function      \n");
    printf("      otp_util -g ofst=<ofst> len=<len> ecc_ofst=<ecc_ofst> ecc_unit=<ecc_unit> \n");
    printf("                                                    \n");

}

static int check_cmd_range(OTP_CMD *p_cmd)
{
    if (p_cmd->msb < p_cmd->lsb) {
        printf("check command range failed, msb (%ld) < lsb (%ld)\n", p_cmd->msb, p_cmd->lsb);
        return -1;
    }

    if (p_cmd->msb > 31) {
        printf("check command range failed, msb (%ld) >= 31\n", p_cmd->msb);
        return -1;
    }

    if (p_cmd->lsb > 31) {
        printf("check command range failed, lsb (%ld) >= 31\n", p_cmd->lsb);
        return -1;
    }

    return 0;
}

/*******************************************************************************
 * FUNC : do_command
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static int do_command(OTP_CMD *p_cmd)
{
    switch (p_cmd->id) {
        case OTP_CMD_DUMP:
            if (p_cmd->len <= 0) {
                if (check_cmd_range(p_cmd) < 0)
                    return 0;
                do_otp_dump(p_cmd->ofst, p_cmd->msb, p_cmd->lsb);
            } else {
                do_otp_read_region(p_cmd->ofst, p_cmd->len);
            }
            break;
        case OTP_CMD_PROGRAM:
            if (check_cmd_range(p_cmd) < 0)
                return 0;
            if (p_cmd->data) {
                do_otp_program_region(p_cmd->ofst, p_cmd->data, p_cmd->len);
            } else {
                if ((p_cmd->flags & VALUE_DETECTED) == 0) {
                    printf("program otp failed, no val specified\n");
                    return 0;
                }
                do_otp_program(p_cmd->ofst, p_cmd->msb, p_cmd->lsb, p_cmd->val);
            }
            break;
        case OTP_CMD_BITMAP_PROGRAM:
            if(p_cmd->data) {
                int size;
                if((size = down_bitmap_file(p_cmd->data)) <= 0)
                    return 0;

                do_otp_program_region(p_cmd->ofst, (unsigned int *)DOWNLOAD_ADDR, size / sizeof(unsigned int));
            } else {
                do_otp_program_with_bitmap();
            }
            break;
        case OTP_CMD_CRC_AUTO_REGION:
            do_otp_crc_auto_region();
            break;
        case OTP_CMD_CRC_PARTIAL_REGION:
            do_otp_crc_partial_region(p_cmd->start_ofst, p_cmd->len);
            break;
        case OTP_CMD_QUERY_EVER_PROGRAMED:
            do_query_otp_ever_programed();
            break;

        case OTP_CMD_ECC_CHECK:
            do_otp_get_data_with_ecc_check(p_cmd->ofst,
                                           p_cmd->len, p_cmd->ecc_ofst,
                                           p_cmd->ecc_unit, 1);
            break;
        case OTP_CMD_HELP:
            do_help();
            break;
    }
    return 0;
}

/*******************************************************************************
 * FUNC : do_command
 * DESC :
 * PARM :
 * RET  :
 *******************************************************************************/
static int parse_command(unsigned char idx, char* str, OTP_CMD *p_cmd)
{
    if (idx == 1) {
        if (strncmp(str, "-d", 2) == 0)
            p_cmd->id = OTP_CMD_DUMP;
        else if (strncmp(str, "-p", 2) == 0)
            p_cmd->id = OTP_CMD_PROGRAM;
        else if(strncmp(str, "-b", 2) == 0)
            p_cmd->id = OTP_CMD_BITMAP_PROGRAM;
        else if (strncmp(str, "-c", 2) == 0)
            p_cmd->id = OTP_CMD_CRC_AUTO_REGION;
        else if (strncmp(str, "-e", 2) == 0)
            p_cmd->id = OTP_CMD_CRC_PARTIAL_REGION;
        else if (strncmp(str, "-h", 2) == 0)
            p_cmd->id = OTP_CMD_HELP;
        else if(strncmp(str, "-q", 2) == 0)
            p_cmd->id = OTP_CMD_QUERY_EVER_PROGRAMED;
        else if(strncmp(str, "-g", 2) == 0)
            p_cmd->id = OTP_CMD_ECC_CHECK;
        else
            return -1;
    } else {
        if (strncmp(str, "ofst=", 5) == 0) {
            if (str_to_long(str + 5, &(p_cmd->ofst), 16) == -1) {
                printf("invalid ofst - %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "bits=", 5) == 0) {
            char buf[4] = {0};
            char *tmp1 = NULL;
            char *tmp2 = NULL;
            tmp1 = str + 5;
            strncpy(buf, tmp1, 2);
            tmp2 = strstr(tmp1, ":");
            if(tmp2 != NULL)
                tmp2 += 1;
            tmp1 = strstr(buf, ":");
            if(tmp1 != NULL)
                tmp1[0] = 0;
            if (str_to_long(buf, &(p_cmd->msb), 10) == -1 || (tmp2 != NULL && str_to_long(tmp2, &p_cmd->lsb, 10) == -1)) {
                printf("invalid bits - %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "val=", 4) == 0) {
            if (str_to_long(str + 4, &(p_cmd->val), 16) == -1) {
                printf("invalid val - %s\n", str);
                return -1;
            }
            p_cmd->flags |= VALUE_DETECTED;
        } else if (strncmp(str, "data=", 5) == 0) {
            if (str_to_long(str + 5,(unsigned long *) &(p_cmd->data), 16) == -1) {
                printf("invalid data addr- %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "start_ofst=", 11) == 0) {
            if (str_to_long(str + 11, &p_cmd->start_ofst, 16) == -1) {
                printf("invalid start_ofst - %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "end_ofst=", 9) == 0) {
            if (str_to_long(str + 9, &(p_cmd->end_ofst), 16) == -1) {
                printf("invalid end_ofst - %s\n", str);
                return -1;
            }
            p_cmd->len = p_cmd->end_ofst - p_cmd->start_ofst + 1;
        } else if (strncmp(str, "len=", 4) == 0) {
            if (str_to_long(str + 4,(unsigned long *) &(p_cmd->len), 10) == -1) {
                printf("invalid len - %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "bitmap_file=", 12) == 0) {
            p_cmd->data = str + 12;
        } else if (strncmp(str, "ecc_ofst=", 9) == 0) {
            if (str_to_long(str + 9, &(p_cmd->ecc_ofst), 16) == -1) {
                printf("invalid ecc_ofst - %s\n", str);
                return -1;
            }
        } else if (strncmp(str, "ecc_unit=", 9) == 0) {
            if (str_to_long(str + 9, &(p_cmd->ecc_unit), 10) == -1) {
                printf("invalid ecc_unit - %s\n", str);
                return -1;
            }
        } else {
            printf("unknown parameter - %s\n", str);
            return -1;
        }
    }
    return 0;
}

static int do_otp(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
    OTP_CMD cmd = {
        .id    = OTP_CMD_HELP,
        .flags = 0,
        .msb   = 31,
        .lsb   = 0,
        .val   = 0,
        .start_ofst = 0,
        .end_ofst = 0,
        .len = 0,
        .ecc_ofst = 0,
        .ecc_unit = 0,
        .data = NULL,
    };
    int i;
    if (argc > 1) {
        for (i = 1; i < argc ; i++) {
            if (parse_command(i, argv[i], &cmd) < 0) {
                cmd.id = OTP_CMD_HELP;
                break;
            }
        }
    }
    do_command(&cmd);
    return 1;
}
U_BOOT_CMD(otp_util, 6, 0, do_otp,
           "Realtek OTP Tool",
           "Format1 : otp_util <opcode> [ofst=<ofst>]bits=<msb>:<lsb>[val=<otp value>][data=<data addr>][len=<len>]\n"
           "Format2 : otp_util <opcode> start_ofst=<start_ofst>[end_ofst=<end_ofst>][val=<otp value>][len=<len>]\n"
           "<opcode> : operations     \n"
           "  -d  : dump otp              \n"
           "  -p  : program otp          \n"
           "  -b  : program otp with otp_map.bin          \n"
           "  -c  : crc auto region       \n"
           "  -e  : crc partial region    \n"
           "  -h  : print help              \n"
           "                                     \n"
           "  <ofst> : otp offset in byte, should be 4 byte align)\n"
           "            presented in hex format, and should be multiple of 4\n"
           "   <msb>  : most significant bits (0~31,default: 31)\n"
           "   <lsb>  : last significant bits (0~31,default: 0) \n"
           "   <bitmap file>  : file that contrain otp bits     \n"
           "   <start_ofst> :the offset of the start reg of the reg region you want to do crc verify     \n"
           "   <end_ofst> :the offset of the last reg of the reg region you want to do crc verify     \n"
           "   <len> :the number of regs of the reg region you want to do crc verify     \n"
           "                                     \n"
           " Example: Dump OTP bits                             \n"
           "      otp_util -d ofst=<ofst> bits=<msb>:<lsb>\n"
           " Example: Dump OTP region                             \n"
           "      otp_util -d ofst=<ofst> len=<len>\n"
           " Example: Program OTP bits (direct assign value)    \n"
           "      otp_util -p ofst=<ofst> bits=<msb>:<lsb> val=<val (MSB first)> \n"
           " Example: Program OTP bits (the data should be contained in DDR)      \n"
           "      otp_util -p ofst=<ofst> data=<data addr> len=<len>      \n"
           " Example: Program OTP bits with default otp_map.bin      \n"
           "      otp_util -b      \n"
           " Example: Program OTP bits use bitmap file downloaded from tftp server      \n"
           "               Note:before run this command,plese set tftp parameters.     \n"
           "      otp_util -b [ofst=<ofst>] bitmap_file=<bitmap_file>       \n"
           " Example: crc auto region      \n"
           "      otp_util -c       \n"
           " Example: crc partial region      \n"
           "      otp_util -e start_ofst=<ofst> end_ofst=<ofst>      \n"
           " Example: crc partial region      \n"
           "      otp_util -e start_ofst=<ofst> len=<len>      \n"
           " Example: query whether OTP has ever been programed      \n"
           "      otp_util -q      \n"
           " Example: ecc check function      \n"
           "      otp_util -g ofst=<ofst> len=<len> ecc_ofst=<ecc_ofst> ecc_unit=<ecc_unit> \n"
          );
