/*
 *
 *  rtk_mcp_hw.c
 *
 *  Implementation of hardware CP module
 *
 *
 **/

#include <common.h>
#include <asm/arch/system.h>
#include <asm/arch/error_type.h>
#include <drivers/mcp/rtk_mcp_hw.h>
#include <drivers/mcp/rtk_mcp_core.h>

extern unsigned int alloc_mcp_descriptor_buffer(bool cache_type);
extern void free_mcp_descriptor_buffer(unsigned int alloc_mem, bool cache_type);
extern t_mcp_descriptor *alloc_mcp_descriptor(void);

#ifdef MCP_DEBUG
static void rtk_mcp_dump_descriptor(t_mcp_descriptor* d)
{
    if(d == NULL)
        return;
    MCP_DBG("\n\nXXXXXXXX desc (%p) XXXXXXXXXX\n", d);
    MCP_DBG("mode=%08x\n", d->mode);
    MCP_DBG("Key0=%08x\n", d->key[0]);
    MCP_DBG("Key1=%08x\n", d->key[1]);
    MCP_DBG("Key2=%08x\n", d->key[2]);
    MCP_DBG("Key3=%08x\n", d->key[3]);
    MCP_DBG("Key4=%08x\n", d->key[4]);
    MCP_DBG("Key5=%08x\n", d->key[5]);
    MCP_DBG("IV0=%08x\n", d->ini_key[0]);
    MCP_DBG("IV1=%08x\n", d->ini_key[1]);
    MCP_DBG("IV2=%08x\n", d->ini_key[2]);
    MCP_DBG("IV3=%08x\n", d->ini_key[3]);
    MCP_DBG("di=%08x\n", d->src_addr);
    MCP_DBG("do=%08x\n", d->dst_addr);
    MCP_DBG("len=%08x\n", d->length);
    MCP_DBG("XXXXXXXX desc end XXXXXXXXXX\n\n");
}
#endif
static int rtk_mcp_do_hw_crypt(t_mcp_descriptor *dscpt)
{
    int res;
    int count;
    if (dscpt == NULL)
        return -1;
    
    DCACHE_FLUSH();

    // disable interrupt
    WRITE_REG32(K_MCP_EN, 0xfe);
    // disable go bit
    WRITE_REG32(K_MCP_CTRL, 0x2);

    // set ring buffer register
    WRITE_REG32(K_MCP_BASE, PHYS(((unsigned int) dscpt)) );
    WRITE_REG32(K_MCP_LIMIT, PHYS(((unsigned int) dscpt + sizeof(t_mcp_descriptor) * 2) ));
    WRITE_REG32(K_MCP_RDPTR, PHYS(((unsigned int) dscpt)));
    WRITE_REG32(K_MCP_WRPTR, PHYS(((unsigned int) dscpt + sizeof(t_mcp_descriptor)) )); // writer pointer cannot be equal to limit
    

#ifdef MCP_DEBUG
    rtk_mcp_dump_descriptor(dscpt);
    MCP_DBG("\nMCP ring buffer registers:\n");
    MCP_DBG("\tMCP_BASE: 0x%x\n\tMCP_LIMIT: 0x%x\n\tMCP_RDPTR: 0x%x\n\tMCP_WRPTR: 0x%x\n",
    READ_REG32(K_MCP_BASE), READ_REG32(K_MCP_LIMIT), READ_REG32(K_MCP_RDPTR), READ_REG32(K_MCP_WRPTR));
#endif

    SYNC();

#if defined(CONFIG_KCPU_MCP)
    // swap descriptor content (KCPU is little-endian)
    WRITE_REG32(K_MCP_CTRL, 0x9);
#endif

    // go
    WRITE_REG32(K_MCP_CTRL, 0x3);

    count = 0;
    while ((READ_REG32(K_MCP_STATUS) & 0x6) == 0) {
        count++;
        if (count == 0x800000) {
#ifdef MCP_DEBUG
            MCP_DBG("%s timeout\n", __FUNCTION__);
            MCP_DBG("REG32(MCP_CTRL): 0x%x\n", READ_REG32(K_MCP_CTRL));
            MCP_DBG("REG32(MCP_STATUS): 0x%x\n", READ_REG32(K_MCP_STATUS));
#endif

            return -3;
        }
    }

#ifdef MCP_DEBUG
    MCP_DBG("mcp done counter: 0x%x\n", count);
    MCP_DBG("MCP status: 0x%x\n", READ_REG32(K_MCP_STATUS));
#endif

    // check result
    if (READ_REG32(K_MCP_STATUS) & 0x2)
        res = 0;        // ring buffer is empty
    else if (READ_REG32(K_MCP_STATUS) & 0x4)
        res = -1;       // error happened
    else
        res = -2;       // ???

    // clear MCP register
    WRITE_REG32(K_MCP_CTRL, 0x2);
    WRITE_REG32(K_MCP_STATUS, 0x6);

    DCACHE_FLUSH_INVALIDATE();

    return res;
}


int reload_otp(void)
{
    unsigned int i;

    WRITE_REG32(CP_OTP_LOAD, 0x1);
    for( i = 0; i < 0x10000; i++ ) {
        if ((READ_REG32(CP_OTP_LOAD) & 0x1) == 0)
            return 0;
    }
    MCP_DBG("reload %s timeout\n", __FUNCTION__);
    return -1;
}

static int load_otp(void)
{
    static int flag = 0;

    if (flag == 0) {
        if(reload_otp() == 0)
            flag = 1;
    } else {
        return 0;
    }
    return flag;
}

void update_iv(t_mcp_descriptor *dscpt, unsigned char * vector, bool is_big_endian)
{
    if ((dscpt == NULL) || (vector == NULL))
        return;
    if(is_big_endian) {
        dscpt->ini_key[0] = MCP_B4TL(vector[0], vector[1], vector[2], vector[3]);
        dscpt->ini_key[1] = MCP_B4TL(vector[4], vector[5], vector[6], vector[7]);
        dscpt->ini_key[2] = MCP_B4TL(vector[8], vector[9], vector[10], vector[11]);
        dscpt->ini_key[3] = MCP_B4TL(vector[12], vector[13], vector[14], vector[15]);
    } else {
        dscpt->ini_key[0] = MCP_B4TL(vector[3], vector[2], vector[1], vector[0]);
        dscpt->ini_key[1] = MCP_B4TL(vector[7], vector[6], vector[5], vector[4]);
        dscpt->ini_key[2] = MCP_B4TL(vector[11], vector[10], vector[9], vector[8]);
        dscpt->ini_key[3] = MCP_B4TL(vector[15], vector[14], vector[13], vector[12]);
    }
}

void update_key(t_mcp_descriptor *dscpt, unsigned char * vector)
{
    if ((dscpt == NULL) || (vector == NULL))
        return;
    if((dscpt->mode & 0x1F) == MCP_MODE_DES) {
        dscpt->key[0] = MCP_B4TL(0, vector[0], vector[1], vector[2]);
        dscpt->key[1] = MCP_B4TL(vector[3], vector[4], vector[5], vector[6]);
    } else if((dscpt->mode & 0x1F) == MCP_MODE_3DES) {
        if((dscpt->mode & BIT(5)) == MCP_ENCRYPT) {

            // key A
            dscpt->key[0] = MCP_B4TL(0, vector[0], vector[1], vector[2]);
            dscpt->key[1] = MCP_B4TL(vector[3], vector[4], vector[5], vector[6]);

            // key B
            dscpt->key[2] = MCP_B4TL(0, vector[7], vector[8], vector[9]);
            dscpt->key[3] = MCP_B4TL(vector[10], vector[11], vector[12], vector[13]);

            // key C
            dscpt->key[4] = MCP_B4TL(0, vector[14], vector[15], vector[16]);
            dscpt->key[5] = MCP_B4TL(vector[17], vector[18], vector[19], vector[20]);
        } else {

            // key C
            dscpt->key[4] = MCP_B4TL(0, vector[0], vector[1], vector[2]);
            dscpt->key[5] = MCP_B4TL(vector[3], vector[4], vector[5], vector[6]);

            //key B
            dscpt->key[2] = MCP_B4TL(0, vector[7], vector[8], vector[9]);
            dscpt->key[3] = MCP_B4TL(vector[10], vector[11], vector[12], vector[13]);

            // key A
            dscpt->key[0] = MCP_B4TL(0, vector[14], vector[15], vector[16]);
            dscpt->key[1] = MCP_B4TL(vector[17], vector[18], vector[19], vector[20]);
        }
    } else {
        dscpt->key[0] = MCP_B4TL(vector[3], vector[2], vector[1], vector[0]);
        dscpt->key[1] = MCP_B4TL(vector[7], vector[6], vector[5], vector[4]);
        dscpt->key[2] = MCP_B4TL(vector[11], vector[10], vector[9], vector[8]);
        dscpt->key[3] = MCP_B4TL(vector[15], vector[14], vector[13], vector[12]);
    }

}

int MCP_DESCRIPT_INIT(t_mcp_descriptor *dscpt, unsigned int mcp_mode, unsigned int mcp_bcd, unsigned int mcp_ende, unsigned int mcp_key_sel, unsigned int key[6], unsigned int ini_key[4])
{

    if (dscpt == NULL)
        return -1;

    dscpt->mode = (mcp_mode | mcp_bcd | mcp_ende | mcp_key_sel);
    // Fill key
    if(mcp_key_sel == MCP_KEY_FROM_DSCPT) {
        update_key(dscpt, (unsigned char *)key);
    } else if(mcp_key_sel == MCP_KEY_FROM_OTP) {
        if (load_otp())
            return -2;

    } else if(mcp_key_sel == MCP_KEY_FROM_CW) {
        dscpt->key[0] = REG_KEY_CW_ENTRY((unsigned int)key - 1);
        dscpt->key[1] = dscpt->key[2] = dscpt->key[3] = 0;

    } else if(mcp_key_sel == MCP_KEY_FROM_DDR) {
        dscpt->key[0]  = PHYS((unsigned int) key);
        dscpt->key[1]  = dscpt->key[2] = dscpt->key[3] = dscpt->key[4] = dscpt->key[5] = 0;
    }
    // Fill ini_key
    if((mcp_mode == MCP_MODE_AES) || (mcp_mode == MCP_MODE_AES_192) || (mcp_mode == MCP_MODE_AES_256) || (mcp_mode == MCP_MODE_DES) ||
       (mcp_mode == MCP_MODE_3DES) || (mcp_mode == MCP_MODE_CAM) || (mcp_mode == MCP_MODE_CAM_192) || (mcp_mode == MCP_MODE_CAM_256)) {
        if(ini_key != NULL && (unsigned int)ini_key <= MCP_MAX_CW_ENTRY) {
            dscpt->mode |= MCP_IV_FROM_CW;
            dscpt->ini_key[0] = REG_IV_CW_ENTRY((unsigned int)ini_key - 1);
            dscpt->ini_key[1] = dscpt->ini_key[2] = dscpt->ini_key[3] = 0;
        } else if(ini_key != NULL) {
            update_iv(dscpt, (unsigned char *)ini_key, 0);
        }
    } else if(mcp_mode == MCP_MODE_SHA256) {
        if (ini_key == NULL) {
            dscpt->key[0] = SHA256_H0;
            dscpt->key[1] = SHA256_H1;
            dscpt->key[2] = SHA256_H2;
            dscpt->key[3] = SHA256_H3;
            dscpt->key[4] = SHA256_H4;
            dscpt->key[5] = SHA256_H5;
            dscpt->ini_key[0] = SHA256_H6;
            dscpt->ini_key[1] = SHA256_H7;
        } else {
            dscpt->key[0] = key[0];
            dscpt->key[1] = key[1];
            dscpt->key[2] = key[2];
            dscpt->key[3] = key[3];
            dscpt->key[4] = ini_key[0];
            dscpt->key[5] = ini_key[1];
            dscpt->ini_key[0] = ini_key[2];
            dscpt->ini_key[1] = ini_key[3];
        }
    } else if(mcp_mode == MCP_MODE_SHA1) {
        if ((key == NULL) || (ini_key == NULL)) {
            dscpt->key[0] = SHA1_IV_0;
            dscpt->key[1] = SHA1_IV_1;
            dscpt->key[2] = SHA1_IV_2;
            dscpt->key[3] = SHA1_IV_3;
            dscpt->key[4] = SHA1_IV_4;
        } else {
            dscpt->key[0] = key[0];
            dscpt->key[1] = key[1];
            dscpt->key[2] = key[2];
            dscpt->key[3] = key[3];
            dscpt->key[4] = ini_key[0];
        }
    } else if(mcp_mode == MCP_MODE_AES_H) {
        if (ini_key == NULL) {
            dscpt->ini_key[0] = AES_H_IV_0;
            dscpt->ini_key[1] = AES_H_IV_1;
            dscpt->ini_key[2] = AES_H_IV_2;
            dscpt->ini_key[3] = AES_H_IV_3;
        } else {
            dscpt->ini_key[0] = ini_key[0];
            dscpt->ini_key[1] = ini_key[1];
            dscpt->ini_key[2] = ini_key[2];
            dscpt->ini_key[3] = ini_key[3];
        }
    }
    return 0;
}

int GenerateMCPDescriptor(unsigned int mcp_mode, unsigned int mcp_bcd, unsigned int mcp_ende, unsigned int mcp_key_sel, unsigned char *src_addr, unsigned int length, unsigned char *dst_addr, unsigned int key[6], unsigned int ini_key[4])
{
    int ret = 0;
    t_mcp_descriptor *dscpt;
    unsigned int alloc_mem;
    if ((src_addr == NULL) || (dst_addr == NULL)) {
        return ERR_INVALID_PARAM;
    }

    alloc_mem = alloc_mcp_descriptor_buffer(DEFUALT_MEM_ALLOC_TYPE);
    dscpt = alloc_mcp_descriptor();
    if (dscpt == NULL) {
        free_mcp_descriptor_buffer(alloc_mem, DEFUALT_MEM_ALLOC_TYPE);
        return ERR_ALLOC_FAILED;
    }

    ret = MCP_DESCRIPT_INIT(dscpt, mcp_mode, mcp_bcd, mcp_ende, mcp_key_sel, key, ini_key);
    if(ret) {
        free_mcp_descriptor_buffer(alloc_mem, DEFUALT_MEM_ALLOC_TYPE);
        return ret;
    }

    dscpt->src_addr = (unsigned int) src_addr;
    dscpt->dst_addr = (unsigned int) dst_addr;
    dscpt->length = length;

    ret = rtk_mcp_do_hw_crypt(dscpt);
    free_mcp_descriptor_buffer(alloc_mem, DEFUALT_MEM_ALLOC_TYPE);
    return ret;
}



